#ifndef _DWC_EVENT_C
#define _DWC_EVENT_C

#ifndef lint
static const char rcsid[] = "$Id: dwc_event.c,v 1.10 2001/10/15 11:04:12 cvser Exp $";
#endif

#if defined(__THREAD_BASED_DW)

#ifdef __UNIX_DW
#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>
#if defined(EMX_DW)
#include <io.h>
#endif /* EMX_DW */
#endif /* __UNIX_DW */

#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include "dwc_macro.h"
#include "dwc_lib.h"
#include "dwc_err.h"
#include "dwc_event.h"

struct DWC_int_event_t {
    DWC_string_t      name;
#if defined(__UNIX_DW)
    int               pipefds[2];
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    HANDLE            handle;
#endif /* __WIN32_DW */
};


#if defined(__UNIX_DW)
/* -------------------------------------------------------------------- */
static
DWC_bool_t set_nonblock( int fd )
{
    int         flags;

    flags = fcntl( fd, F_GETFL );
    if( flags < 0 )
        return FALSE;
    flags |= O_NONBLOCK;
    if( fcntl( fd, F_SETFL, flags ) < 0 )
        return FALSE;
    else
        return TRUE;
}
#endif /* __UNIX_DW */

/* -------------------------------------------------------------------- */
DWC_event_t dwc_event_create( DWC_string_t name )
{
    int         error;
    DWC_event_t event;

    event = DWC_Malloc( DWC_event_t, sizeof(struct DWC_int_event_t) );
    if( !event )
    {
        if( name )
            dprintLib( "creating of event '%s' failed - no memory.\n", name );
        else
            dprintLib( "creating of anonymous event failed - no memory.\n" );
        return NULL;
    }
    if( name )
        event->name = strdup( name );
    else
        event->name = strdup( "anonymous" );
    if( !event->name )
    {
        DWC_Free( event );
        if( name )
            dprintLib( "creating of event '%s' failed - no memory.\n", name );
        else
            dprintLib( "creating of anonymous event failed - no memory.\n" );
        return NULL;
    }
    error = 0;
#if defined(__UNIX_DW)
    if( pipe( event->pipefds ) != 0 )
    {
        error = errno;
    }
    if( !set_nonblock( event->pipefds[0] )
        ||
        !set_nonblock( event->pipefds[1] ) )
    {
        error = errno;
        close( event->pipefds[0] );
        close( event->pipefds[1] );
    }
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    event->handle = CreateEvent( NULL, TRUE, FALSE, event->name );
    if( !event->handle )
    {
        error = errno;
    }
#endif /* __WIN32_DW */
    if( error != 0 )
    {
        free( event->name );
        DWC_Free( event );
        dprintLib( "creating of event '%s'failed - %s.\n", event->name, stdlib_error( error ) );
        return NULL;
    }
    return event;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t dwc_event_wait( DWC_event_t event, DWC_timeout_t timeout )
{
#if defined(__UNIX_DW)
    struct timeval      timo, *pt;
    fd_set              rd_set;
    int                 rd;
    char                inbuf;
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    DWORD               w_timeout;
#endif /* __WIN32_DW */
    int                 error = 0;
    DWC_error_code_t    retval = DWC_NOERROR;

    PARAM_ASSERT( event );

#if defined(__UNIX_DW)
    FD_ZERO( &rd_set );
    FD_SET( event->pipefds[0], &rd_set );
    if( timeout != INFINITE_TIMEOUT )
    {
        timo.tv_sec = timeout / 1000;
        timo.tv_usec = (timeout % 1000) * 1000;
        pt = &timo;
    }
    else
    {
        pt = NULL;
    }
    rd = select( event->pipefds[0]+1, &rd_set, NULL, NULL, pt );
    if( (rd == 1) && FD_ISSET( event->pipefds[0], &rd_set ) )
    {
        if( read( event->pipefds[0], &inbuf, 1 ) != 1 )
        {
            error = errno;
        }
    }
    else if( (rd == 0) && pt )
    {
        retval = DWC_TIMED;
    }
    else
    {
        error = errno;
    }
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    if( timeout == INFINITE_TIMEOUT )
    {
        w_timeout = INFINITE;
    }
    else
    {
        w_timeout = timeout;
    }
    error = WaitForSingleObject( event, w_timeout );
    if( error == WAIT_OBJECT_0 )
    {
        error = 0;
        retval = DWC_NOERROR;
    }
    else if( error == WAIT_TIMEOUT )
    {
        error = 0;
        retval = DWC_TIMED;
    }
    else
    {
        error = errno;
    }
#endif /* __WIN32_DW */
    if( error )
    {
        dprintLib( "waiting for event '%s' failed - %s.\n", event->name, stdlib_error( error ) );
        retval = DWC_COMMONERROR;
    }
    return retval;
}

/* -------------------------------------------------------------------- */
int dwc_event_multiple_wait( DWC_event_t* events, DWC_length_t num, DWC_timeout_t timeout )
{
#if defined(__UNIX_DW)
    struct timeval      timo, *pt;
    fd_set              rd_set;
    int                 rd,maxfd;
    char                inbuf;
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    HANDLE             *hnd_array;
    DWORD               w_timeout;
#endif /* __WIN32_DW */
    int                 i;
    int                 error = 0;
    int                 retval = DWC_COMMONERROR;

    PARAM_ASSERT( events );
    PARAM_ASSERT( num > 0 );

#if defined(__UNIX_DW)
    FD_ZERO( &rd_set );
    for( i = 0, maxfd = 0; i < num; ++i )
    {
        FD_SET( events[i]->pipefds[0], &rd_set );
        if( events[i]->pipefds[0] > maxfd )
            maxfd = events[i]->pipefds[0];
    }
    if( timeout != INFINITE_TIMEOUT )
    {
        timo.tv_sec = timeout / 1000;
        timo.tv_usec = (timeout % 1000) * 1000;
        pt = &timo;
    }
    else
    {
        pt = NULL;
    }
    rd = select( maxfd+1, &rd_set, NULL, NULL, pt );
    if( rd > 0 )
    {
        for( i = 0; i < num; ++i )
        {
            if( FD_ISSET( events[i]->pipefds[0], &rd_set ) )
            {
                if( (read( events[i]->pipefds[0], &inbuf, 1 ) == 1) )
                {
                    retval = i;
                }
                else
                {
                    error = errno;
                }
                break;
            }
        }
    }
    else if( (rd == 0) && pt )
    {
        retval = num;
    }
    else
    {
        error = errno;
    }
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    hnd_array = DWC_Malloc( HANDLE*, num * sizeof(HANDLE) );
    if( !hnd_array )
    {
        dprintLib("waiting for multiple events failed - no memory.\n");
        return DWC_COMMONERROR;
    }
    for( i = 0; i < num; ++i )
    {
        hnd_array[i] = events[i]->handle;
    }
    if( timeout == INFINITE_TIMEOUT )
    {
        w_timeout = INFINITE;
    }
    else
    {
        w_timeout = timeout;
    }
    error = WaitForMultipleObjects( num, hnd_array, FALSE, w_timeout );
    if( error == WAIT_TIMEOUT )
    {
        retval = num;
        error = 0;
    }
    else if( (WAIT_OBJECT_0 <= error) && (error < WAIT_OBJECT_0 + num) )
    {
        retval = error - WAIT_OBJECT_0;
        error = 0;
    }
    else
    {
        error = errno;
    }
    DWC_Free( hnd_array );
#endif /* __WIN32_DW */
    if( error )
    {
        dprintLib( "waiting for multiple events failed - %s.\n", stdlib_error( error ) );
        retval = DWC_COMMONERROR;
    }
    return retval;
}

/* -------------------------------------------------------------------- */
DWC_bool_t dwc_event_set( DWC_event_t event )
{
    int         error;
    char        outbuf = 'E';

    PARAM_ASSERT( event );

    error = 0;
#if defined(__UNIX_DW)
    if( write( event->pipefds[1], &outbuf, 1 ) != 1 )
    {
        dprintLib( "writing to pipe failed (%p:%d).\n", event, event->pipefds[1] );
        error = errno;
    }
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    if( !SetEvent( event->handle ) )
    {
        error = errno;
    }
#endif /* __WIN32_DW */
    if( error != 0 )
    {
        dprintLib( "setting of event '%s' failed - %s.\n", event->name, stdlib_error( error ) );
        return FALSE;
    }
    return TRUE;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t dwc_event_destroy( DWC_event_t event )
{
    DWC_error_code_t    retcode;
    int                 error;

    PARAM_ASSERT( event );

    error = 0;
#if defined(__UNIX_DW)
    if( close( event->pipefds[0] ) != 0 )
    {
        error = errno;
    }
    if( close( event->pipefds[1] ) != 0 )
    {
        error = errno;
    }
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    if( !CloseHandle( event->handle ) )
    {
        error = errno;
    }
#endif /* __WIN32_DW */
    if( error != 0 )
    {
        dprintLib( "destroying of event '%s' failed - %s.\n", event->name, stdlib_error( error ) );
        retcode = DWC_COMMONERROR;
    }
    else
    {
        retcode = DWC_NOERROR;
    }
    free( event->name );
    DWC_Free( event );
    return retcode;
}

#endif /* __THREAD_BASED_DW */

#endif /* _DWC_EVENT_C */
