#ifndef _DWC_IPC_C
#define _DWC_IPC_C

#ifndef lint
static const char rcsid[] = "$Id: dwc_ipc.c,v 1.29 2001/10/08 15:48:52 cvser Exp $";
#endif

#include <errno.h>
#include <string.h>
#if defined(__UNIX_DW)
#if defined(FreeBSD_DW)
#include <sys/time.h>
#include <unistd.h>
#endif /* FreeBSD */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#endif /* __UNIX_DW */

#include "dwc_lib.h"
#include "dwc_err.h"
#include "dwc_macro.h"
#include "dwc_timo.h"
#include "dwc_ipc.h"
#include "dwc_tcp.h"
#include "dwc_unix.h"

typedef union {
    DWC_tcp_socket_t        hnd_tcp;
#if defined(__UNIX_DW)
    DWC_unix_socket_t       hnd_unix;
#endif
#if (defined(FreeBSD_DW) && defined(__THREAD_BASED_DW)) || defined(Solaris_DW)
    int                     fd;
#endif /* FreeBSD */
} DWC_ipc_handle_t;

typedef union {
    DWC_tcp_socket_addr_t   sa_tcp;
#if defined(__UNIX_DW)
    DWC_unix_socket_addr_t  sa_unix;
#endif
} DWC_ipc_address_t;

#define DW_MAX_ADDR_LEN     (256)

typedef struct _DWC_ipc_addr {
    DWC_char_t              name[DW_MAX_ADDR_LEN];
    DWC_length_t            name_len;
    DWC_ipc_address_t       addr;
} DWC_ipc_addr_t;

#define NULL_IPC_ADDR( a_ipc )\
BEGIN_MACRO\
    (a_ipc).name[0] = NULLCHAR;\
    (a_ipc).name_len = 0;\
    memset( &(a_ipc).addr, 0, sizeof(DWC_ipc_address_t) );\
END_MACRO

typedef DWC_error_code_t   (*DWC_SendFunc_t)();
typedef DWC_error_code_t   (*DWC_RecvFunc_t)();
typedef DWC_error_code_t   (*DWC_WaitFunc_t)();

struct _DWC_ipc {
    DWC_ipc_type_t          type;
    DWC_bool_t              allocated;
    DWC_bool_t              connected;
    DWC_bool_t              nonblocking;
    DWC_bool_t              listed;
    DWC_bool_t              binded;
    DWC_ipc_handle_t        handle;
#if defined(FreeBSD_DW) && defined(__THREAD_BASED_DW)
    int                     closer[2];
#endif /* FreeBSD */
    DWC_ipc_addr_t          local;
    DWC_ipc_addr_t          remote;
    DWC_SendFunc_t          send_func;
    DWC_RecvFunc_t          recv_func;
    DWC_WaitFunc_t          wait_func;
    DWC_TimeoutProc_t       cancel_call;
};

typedef struct {
    unsigned int            inited;
    unsigned int            tcp_sock:1,
                            unix_sock:1;
#if defined(__UNIX_DW)
    unsigned int            act_saved;
    struct sigaction        old_act;
#endif /* __UNIX_DW */
} DWC_ipc_subsys_state_t;

static DWC_ipc_subsys_state_t ipcState = { 0 };

/* -------------------------------------------------- */
DWC_ipc_addr_ptr_t create_ipc_addr ( DWC_ipc_type_t ipc_type, DWC_string_t name, ... )
{
    DWC_ipc_addr_ptr_t  addr;
    DWC_error_code_t    error;
    va_list             args;

    PARAM_ASSERT( name );

    addr = DWC_Malloc( DWC_ipc_addr_ptr_t, sizeof(*addr) );
    if( !addr )
    {
        dprintLib( "IPC: no memory for address.\n" );
        return NULL;
    }
    memset( addr, 0, sizeof(*addr) );
    switch( ipc_type )
    {
        case DWC_IPC_TCP_SOCKET:
            va_start( args, name );
            error = addr_tcp_op( &addr->addr.sa_tcp, DWC_IPC_CMD_RESOLVE, name, args );
            if( error == DWC_NOERROR )
            {
                error = addr_tcp_op( &addr->addr.sa_tcp, DWC_IPC_CMD_SETPORT, name, args );
                addr->addr.sa_tcp.sin_family = PF_INET;
            }
            va_end( args );
            break;
#if defined(__UNIX_DW)
        case DWC_IPC_UNIX_SOCKET:
            error = addr_unix_op( &addr->addr.sa_unix, DWC_IPC_CMD_SETADDR, name );
            addr->addr.sa_unix.sun_family = PF_UNIX;
            break;
#endif /* __UNIX_DW */
        default:
            error = DWC_BADIPC;
            break;
    }
    if( error != DWC_NOERROR )
    {
        dprintLib( "IPC: caching failed.\n" );
        DWC_Free( addr ); /* addr == NULL after that */
    }
    else
    {
        strncpy( addr->name, name, sizeof(addr->name) );
        addr->name_len = DWC_strlen( addr->name );
        if( ipc_type == DWC_IPC_TCP_SOCKET )
        {
            dprintLib("IPC: cache address %d@%s:%d for [%d]:%s\n", addr->addr.sa_tcp.sin_family, inet_ntoa( addr->addr.sa_tcp.sin_addr ), ntohs(addr->addr.sa_tcp.sin_port), addr->name_len, addr->name );
        }
    }
    return addr;
}

/* -------------------------------------------------- */
DWC_error_code_t copy_ipc_addr ( DWC_ipc_t ipc, DWC_side_t side, DWC_ipc_addr_ptr_t addr )
{
    DWC_ipc_addr_ptr_t      naddr;

    PARAM_ASSERT( ipc );
    PARAM_ASSERT( addr );

    switch( side )
    {
        case DWC_IPC_SIDE_LOCAL:
            naddr = &ipc->local;
            break;
        case DWC_IPC_SIDE_REMOTE:
            naddr = &ipc->remote;
            break;
        default:
            return DWC_BADPARAMS;
    }
    memcpy( naddr, addr, sizeof(*naddr) );
    return DWC_NOERROR;
}

/* -------------------------------------------------- */
DWC_error_code_t delete_ipc_addr ( DWC_ipc_addr_ptr_t addr )
{
    PARAM_ASSERT( addr );
    DWC_Free( addr );
    return DWC_NOERROR;
}

/* -------------------------------------------------- */
DWC_error_code_t init_ipc_subsystem ( void )
{
#if defined(__UNIX_DW)
    struct sigaction        ignore;
#endif /* __UNIX_DW */
    int ret;

    if( ipcState.inited == ENABLE )
    {
        dprintLib( "IPC: system has already been initialized.\n" );
        return DWC_ALREADYPERFORM;
    }
    ret = InitTimerEngine();
    if( ret != DWC_NOERROR )
    {
        dprintLib( "IPC: can not initialize timers engine.\n" );
        return DWC_IPCSYSTEM;
    }
#if defined(__UNIX_DW)
    ipcState.act_saved = TRUE;
    ignore.sa_flags = 0;
    ignore.sa_handler = SIG_IGN;
    sigemptyset(&ignore.sa_mask);
    if( sigaction( SIGPIPE, &ignore, &ipcState.old_act ) < 0 )
    {
        dprintLib( "IPC: can not ignore SIGPIPE.\n" );
        ipcState.act_saved = FALSE;
    }
#endif /* __UNIX_DW */
    ret = init_tcp_sockets();
    if( ret == DWC_NOERROR )
    {
        ipcState.tcp_sock = ENABLE;
    }
    else
    {
        ipcState.tcp_sock = DISABLE;
    }
#if defined(__UNIX_DW)
    ret = init_unix_sockets();
    if( ret == DWC_NOERROR )
    {
        ipcState.unix_sock = ENABLE;
    }
    else
    {
        ipcState.unix_sock = DISABLE;
    }
#else /* __UNIX_DW */
    ipcState.unix_sock = DISABLE;
#endif /* __UNIX_DW */
    ipcState.inited = ENABLE;
    return DWC_NOERROR;
}

/* -------------------------------------------------- */
DWC_error_code_t dinit_ipc_subsystem ( void )
{
    int tret = DWC_NOERROR;
    int uret = DWC_NOERROR;
    int eret = DWC_NOERROR;

    if( ipcState.inited != ENABLE )
    {
        dprintLib( "IPC system is not initialized.\n" );
        return DWC_NOTREADY;
    }
    eret = DeInitTimerEngine();
    if( eret != DWC_NOERROR )
    {
        dprintLib( "IPC: can not deinitialize timers engine.\n" );
    }
    if( ipcState.tcp_sock == ENABLE )
    {
        tret = dinit_tcp_sockets();
    }
    else
    {
        tret = DWC_NOERROR;
    }
#if defined(__UNIX_DW)
    if( ipcState.unix_sock == ENABLE )
    {
        uret = dinit_unix_sockets();
    }
    else
    {
        uret = DWC_NOERROR;
    }
    if( !ipcState.act_saved
        ||
        (sigaction( SIGPIPE, &ipcState.old_act, NULL ) < 0) )
    {
        dprintLib( "IPC: can not restore SIGPIPE handler.\n" );
        uret = DWC_COMMONERROR;
    }
#else /* __UNIX_DW */
    uret = DWC_NOERROR;
#endif /* __UNIX_DW */
    return ( eret == DWC_NOERROR ? ( tret == DWC_NOERROR ? uret : tret ) : eret );
}

/* -------------------------------------------------- */
DWC_error_code_t create_ipc ( DWC_ipc_type_t type, DWC_ipc_t *ipc )
{
    DWC_error_code_t error;

    PARAM_ASSERT( ipc );
    PARAM_ASSERT( ((DWC_IPC_TYPE_0 < type) && (type < DWC_IPC_TYPE_MAX)) );

    if( !(*ipc) )
    {
        (*ipc) = DWC_Malloc( DWC_ipc_t, sizeof(struct _DWC_ipc) );
        if( !ipc )
        {
            return DWC_NOMEMORY;
        }
        (*ipc)->allocated = TRUE;
    }
    else
    {
        (*ipc)->allocated = FALSE;
    }
    (*ipc)->type = type;
    (*ipc)->connected = FALSE;
    (*ipc)->binded = FALSE;
#if defined(__WIN32_DW)
    (*ipc)->nonblocking = TRUE;
#else /* __WIN32_DW */
    (*ipc)->nonblocking = FALSE;
#endif /* __WIN32_DW */
    (*ipc)->listed = FALSE;
    NULL_IPC_ADDR( (*ipc)->local );
    NULL_IPC_ADDR( (*ipc)->remote );
    switch( type )
    {
        case DWC_IPC_TCP_SOCKET:
            if( ipcState.tcp_sock != ENABLE )
            {
                error = DWC_BADIPC;
                goto error_exit;
            }
            error = create_tcp_socket( &(*ipc)->handle.hnd_tcp, (*ipc)->nonblocking );
            if( error != DWC_NOERROR )
            {
                goto error_exit;
            }
            (*ipc)->send_func = send_tcp_socket;
            (*ipc)->recv_func = recv_tcp_socket;
            (*ipc)->wait_func = wait_tcp_socket;
            (*ipc)->cancel_call = cancel_tcp_call;
            (*ipc)->local.addr.sa_tcp.sin_family = PF_INET;
            (*ipc)->remote.addr.sa_tcp.sin_family = PF_INET;
            break;
#if defined(__UNIX_DW)
        case DWC_IPC_UNIX_SOCKET:
            if( ipcState.unix_sock != ENABLE )
            {
                error = DWC_BADIPC;
                goto error_exit;
            }
            error = create_unix_socket( &(*ipc)->handle.hnd_unix, (*ipc)->nonblocking );
            if( error != DWC_NOERROR )
            {
                goto error_exit;
            }
            (*ipc)->send_func = send_unix_socket;
            (*ipc)->recv_func = recv_unix_socket;
            (*ipc)->wait_func = wait_unix_socket;
            (*ipc)->cancel_call = cancel_unix_call;
            (*ipc)->local.addr.sa_unix.sun_family = PF_UNIX;
            (*ipc)->remote.addr.sa_unix.sun_family = PF_UNIX;
            break;
#endif /* __UNIX_DW */
        default:
            return DWC_BADIPC;
    }
    return DWC_NOERROR;

  error_exit:
    delete_ipc( ipc );
    return error;
}

/* -------------------------------------------------- */
DWC_error_code_t delete_ipc ( DWC_ipc_t* ipc )
{
    DWC_error_code_t error;

    PARAM_ASSERT( ipc );

    if( !(*ipc) )
        return DWC_BADIPC;
    switch( (*ipc)->type )
    {
        case DWC_IPC_TCP_SOCKET:
            error = close_tcp_socket( &(*ipc)->handle.hnd_tcp );
            break;
#if defined(__UNIX_DW)
        case DWC_IPC_UNIX_SOCKET:
            error = close_unix_socket( &(*ipc)->handle.hnd_unix, (*ipc)->local.addr.sa_unix, (*ipc)->binded );
            break;
#endif /* __UNIX_DW */
        default:
            return DWC_BADIPC;
    }
    if( (*ipc)->allocated )
        DWC_Free( *ipc );
    return DWC_NOERROR;
}

/* -------------------------------------------------- */
DWC_error_code_t addr_ipc_op ( DWC_ipc_t ipc, DWC_cmd_t cmd, DWC_side_t side, DWC_string_t name, DWC_length_t name_len, ... )
{
    DWC_ipc_addr_t      *fulladdr;
    DWC_error_code_t     error;
    va_list              args;

    if( (!name && (name_len != 0))
        ||
        (name && (name_len == 0))
        ||
        (name && (name_len != DWC_strlen(name))) )
    {
        return DWC_BADPARAMS;
    }
    /* name_len include zero-terminator */
    if( name_len > DW_MAX_ADDR_LEN )
    {
        return DWC_TOOLONG;
    }
    switch( side )
    {
        case DWC_IPC_SIDE_LOCAL:
            fulladdr = &ipc->local;
            break;
        case DWC_IPC_SIDE_REMOTE:
            fulladdr = &ipc->remote;
            break;
        default:
            return DWC_BADPARAMS;
    }
    if( name )
    {
        strncpy( &fulladdr->name[0], name, DW_MAX_ADDR_LEN );
        fulladdr->name_len = DWC_strlen(fulladdr->name);
    }
    else
    {
        fulladdr->name[0] = NULLCHAR;
        fulladdr->name_len = 0;
    }
    va_start( args, name_len );
    switch( ipc->type )
    {
        case DWC_IPC_TCP_SOCKET:
            error = addr_tcp_op( &fulladdr->addr.sa_tcp, cmd, fulladdr->name, args );
            break;
#if defined(__UNIX_DW)
        case DWC_IPC_UNIX_SOCKET:
            error = addr_unix_op( &fulladdr->addr.sa_unix, cmd, fulladdr->name );
            break;
#endif /* __UNIX_DW */
        default:
            error = DWC_BADIPC;
            break;
    }
    if( error != DWC_NOERROR )
    {
        goto error_exit;
    }
    error = DWC_NOERROR;
  error_exit:
    va_end( args );
    return error;
}

/* -------------------------------------------------- */
DWC_error_code_t bind_ipc ( DWC_ipc_t ipc )
{
    DWC_error_code_t  error;

    PARAM_ASSERT( ipc );

    if( ipc->binded )
    {
        return DWC_ALREADYPERFORM;
    }
    switch( ipc->type )
    {
        case DWC_IPC_TCP_SOCKET:
            error = bind_tcp_socket( ipc->handle.hnd_tcp, &ipc->local.addr.sa_tcp );
            break;
#if defined(__UNIX_DW)
        case DWC_IPC_UNIX_SOCKET:
            error = bind_unix_socket( ipc->handle.hnd_unix, &ipc->local.addr.sa_unix );
            break;
#endif /* __UNIX_DW */
        default:
            return DWC_BADIPC;
    }
    if( error == DWC_NOERROR )
    {
        ipc->binded = TRUE;
    }
    return error;
}

/* -------------------------------------------------- */
DWC_error_code_t listen_ipc ( DWC_ipc_t ipc, int backlog )
{
    DWC_error_code_t  error;

    PARAM_ASSERT( ipc );
    PARAM_ASSERT( backlog > 0 );

    if( ipc->listed || ipc->connected )
    {
        return DWC_ALREADYPERFORM;
    }
    switch( ipc->type )
    {
        case DWC_IPC_TCP_SOCKET:
            error = listen_tcp_socket( ipc->handle.hnd_tcp, backlog );
            break;
#if defined(__UNIX_DW)
        case DWC_IPC_UNIX_SOCKET:
            error = listen_unix_socket( ipc->handle.hnd_unix, backlog );
            break;
#endif
        default:
            return DWC_BADIPC;
    }
    if( error == DWC_NOERROR )
    {
        ipc->listed = TRUE;
    }
    return error;
}

/* -------------------------------------------------- */
DWC_error_code_t accept_ipc ( DWC_ipc_t ipc, DWC_ipc_t* ipc_ptr, DWC_timeout_t timeout )
{
    DWC_error_code_t  error;

    PARAM_ASSERT( ipc );
    PARAM_ASSERT( ipc_ptr );

    if( !ipc->listed )
    {
        return DWC_NOTREADY;
    }
    error = create_ipc( ipc->type, ipc_ptr );
    if( error != DWC_NOERROR )
    {
        return error;
    }
    switch( ipc->type )
    {
        case DWC_IPC_TCP_SOCKET:
            error = accept_tcp_socket( ipc->handle.hnd_tcp, &(*ipc_ptr)->remote.addr.sa_tcp, &(*ipc_ptr)->handle.hnd_tcp );
            if( error != DWC_NOERROR )
            {
                goto error_exit;
            }
            (*ipc_ptr)->remote.name[0] = NULLCHAR;
            (*ipc_ptr)->remote.name_len = 0;
            break;
#if defined(__UNIX_DW)
        case DWC_IPC_UNIX_SOCKET:
            error = accept_unix_socket( ipc->handle.hnd_unix, &(*ipc_ptr)->remote.addr.sa_unix, &(*ipc_ptr)->handle.hnd_unix );
            if( error != DWC_NOERROR )
            {
                goto error_exit;
            }
            goto error_exit;
#endif
        default:
            error = DWC_BADIPC;
            goto error_exit;
    }
    memcpy( &(*ipc_ptr)->local, &ipc->local, sizeof(DWC_ipc_addr_t) );
    (*ipc_ptr)->connected = TRUE;
    return DWC_NOERROR;

  error_exit:
    delete_ipc( ipc_ptr );
    return error;
}

/* -------------------------------------------------- */
DWC_error_code_t connect_ipc ( DWC_ipc_t ipc, DWC_timeout_t timeout )
{
    DWC_error_code_t  error = DWC_COMMONERROR;

    PARAM_ASSERT( ipc );

    if( ipc->listed )
    {
        return DWC_NOTREADY;
    }
    switch( ipc->type )
    {
        case DWC_IPC_TCP_SOCKET:
            error = connect_tcp_socket( ipc->handle.hnd_tcp, &ipc->remote.addr.sa_tcp, timeout );
            if( error == DWC_TIMED )
            {
                dprintLib("IPC operation: timeout expired.\n");
            }
            break;
#if defined(__UNIX_DW)
        case DWC_IPC_UNIX_SOCKET:
            error = connect_unix_socket( ipc->handle.hnd_unix, &ipc->remote.addr.sa_unix, timeout );
            if( error == DWC_TIMED )
            {
                dprintLib("IPC operation: timeout expired.\n");
            }
            break;
#endif /* __UNIX_DW */
        default:
            error = DWC_BADIPC;
            break;
    }
    if( error == DWC_NOERROR )
    {
        ipc->connected = TRUE;
    }
    return error;
}

/* -------------------------------------------------- */
DWC_error_code_t send_ipc ( DWC_ipc_t ipc, DWC_pointer_t buf, DWC_length_t* len_ptr, DWC_timeout_t timeout )
{
    DWC_error_code_t    error = DWC_NOERROR;
    DWC_length_t        needbytes = (*len_ptr);
    DWC_length_t        sendbytes = 0;
    DWC_timer_hnd_t     tid;
    int                 how;
#if defined(FreeBSD_DW) && defined(__THREAD_BASED_DW)
    int                 max_fd;
    fd_set              wfds, efds;
#endif /* FreeBSD */

    PARAM_ASSERT( ipc );
    PARAM_ASSERT( buf );
    PARAM_ASSERT( len_ptr );
    PARAM_ASSERT( *len_ptr >= 0 );

    if( (*len_ptr) == 0 )
    {
        return DWC_NOERROR;
    }
    if( ipc->nonblocking )
    {
        while( sendbytes < needbytes  )
        {
#ifdef Solaris_DW        
            error = ipc->send_func( ipc->handle.fd, (DWC_pointer_t)((char*)buf + sendbytes), len_ptr, 0 );
#else /* !Solaris_DW */
            error = ipc->send_func( ipc->handle, (DWC_pointer_t)((char*)buf + sendbytes), len_ptr, 0 );
#endif /* Solaris_DW */            
            if( error != DWC_NOERROR )
            {
                dprintLib( "Nonblocking IPC: waiting error.\n" );
                break;
            }
            sendbytes += (*len_ptr);
            if( sendbytes >= needbytes )
            {
                break;
            }
            how = DWC_IPC_WAIT_SEND;
#ifdef Solaris_DW        
            error = ipc->wait_func( ipc->handle.fd, &how, timeout );
#else /* !Solaris_DW */
            error = ipc->wait_func( ipc->handle, &how, timeout );
#endif /* Solaris_DW */            
            if( (error != DWC_NOERROR) || !(how & DWC_IPC_WAIT_SEND) )
            {
                if( error == DWC_TIMED )
                {
                    dprintLib( "Nonblocking IPC: timeout expired.\n" );
                }
                else
                {
                    dprintLib( "Nonblocking IPC: waiting error.\n" );
                }
                break;
            }
            *len_ptr = needbytes - sendbytes;
        }
        *len_ptr = sendbytes;
    }
    else
    {
#if defined(FreeBSD_DW) && defined(__THREAD_BASED_DW)
        if( pipe( ipc->closer ) < 0 )
        {
            dprintLib( "Can not create timeout facility for IPC - [%d] %s.\n", errno, stdlib_error(errno) );
            return DWC_NORESOURCE;
        }
        FD_ZERO( &wfds );
        FD_ZERO( &efds );
        FD_SET( ipc->handle.fd, &wfds );
        FD_SET( ipc->closer[0], &efds );
        max_fd = 1 + DWC_Max( ipc->handle.fd, ipc->closer[0] );
        tid = SetTimeout( timeout, ipc->cancel_call, &ipc->closer[1], FALSE );
#else /* ! FreeBSD */
#ifdef Solaris_DW        
        tid = SetTimeout( timeout, ipc->cancel_call, &ipc->handle.fd, FALSE );
#else /* !Solaris_DW */
        tid = SetTimeout( timeout, ipc->cancel_call, &ipc->handle, FALSE );
#endif /* Solaris_DW */            
#endif /* FreeBSD */
        if( tid == INVALID_TIMER )
        {
            dprintLib( "Can not set timeout for IPC operations.\n" );
            return DWC_NORESOURCE;
        }
        while( sendbytes < needbytes )
        {
#if defined(FreeBSD_DW) && defined(__THREAD_BASED_DW)
            if( select( max_fd, NULL, &wfds, &efds, NULL ) < 0 )
            {
                dprintLib("Sending aborted.\n");
                error = DWC_IPCSYSTEM;
            }
            if( FD_ISSET( ipc->closer[0], &efds ) )
            {
                dprintLib("IPC operation: timeout expired.\n");
                error = DWC_TIMED;
                break;
            }
#endif /* FreeBSD */
#ifdef Solaris_DW        
            error = ipc->send_func( ipc->handle.fd, (DWC_pointer_t)((char*)buf + sendbytes), len_ptr, 0 );
#else /* !Solaris_DW */
            error = ipc->send_func( ipc->handle, (DWC_pointer_t)((char*)buf + sendbytes), len_ptr, 0 );
#endif /* Solaris_DW */            
            if( error != DWC_NOERROR )
            {
                if( ExpiredTimeout( tid ) )
                {
                    dprintLib("IPC operation: timeout expired.\n");
                    error = DWC_TIMED;
                }
                break;
            }
            if( ExpiredTimeout( tid ) )
            {
                dprintLib("IPC operation: timeout expired.\n");
                error = DWC_TIMED;
                break;
            }
            sendbytes += (*len_ptr);
            *len_ptr = needbytes - sendbytes;
        }
        ClearTimeout( tid );
#if defined(FreeBSD_DW) && defined(__THREAD_BASED_DW)
        close( ipc->closer[0] );
        if( ipc->closer[1] >= 0 )
            close( ipc->closer[1] );
#endif /* FreeBSD */
        *len_ptr = sendbytes;
    }
    return error;
}

/* -------------------------------------------------- */
DWC_error_code_t recv_ipc ( DWC_ipc_t ipc, DWC_pointer_t buf, DWC_length_t* len_ptr, DWC_timeout_t timeout )
{
    DWC_error_code_t    error = DWC_NOERROR;
    DWC_length_t        needbytes = (*len_ptr);
    DWC_length_t        recvbytes = 0;
    DWC_timer_hnd_t     tid = INVALID_TIMER;
    int                 how;
#if defined(FreeBSD_DW) && defined(__THREAD_BASED_DW)
    int                 max_fd = 0;
    fd_set              rfds, efds;
#endif /* FreeBSD */

    PARAM_ASSERT( ipc );
    PARAM_ASSERT( buf );
    PARAM_ASSERT( len_ptr );
    PARAM_ASSERT( *len_ptr >= 0 );

    if( (*len_ptr) == 0 )
    {
        return DWC_NOERROR;
    }
    if( !ipc->nonblocking )
    {
#if defined(FreeBSD_DW) && defined(__THREAD_BASED_DW)
        if( pipe( ipc->closer ) < 0 )
        {
            dprintLib( "Can not create timeout facility for IPC - [%d] %s.\n", errno, stdlib_error(errno) );
            return DWC_NORESOURCE;
        }
        FD_ZERO( &rfds );
        FD_ZERO( &efds );
        FD_SET( ipc->handle.fd, &rfds );
        FD_SET( ipc->closer[0], &efds );
        max_fd = 1 + DWC_Max( ipc->handle.fd, ipc->closer[0] );
        tid = SetTimeout( timeout, ipc->cancel_call, &ipc->closer[1], FALSE );
#else /* ! FreeBSD */
#ifdef Solaris_DW        
        tid = SetTimeout( timeout, ipc->cancel_call, &ipc->handle.fd, FALSE );
#else /* !Solaris_DW */
        tid = SetTimeout( timeout, ipc->cancel_call, &ipc->handle, FALSE );
#endif /* Solaris_DW */            
#endif /* FreeBSD */
        if( tid == INVALID_TIMER )
        {
            dprintLib( "Can not set timeout for IPC operations.\n" );
            return DWC_NORESOURCE;
        }
    }
    while( recvbytes < needbytes  )
    {
#if defined(FreeBSD_DW) && defined(__THREAD_BASED_DW)
        if( !ipc->nonblocking )
        {
            if( select( max_fd, &rfds, NULL, &efds, NULL ) < 0 )
            {
                dprintLib("Sending aborted.\n");
                error = DWC_IPCSYSTEM;
            }
            if( FD_ISSET( ipc->closer[0], &efds ) )
            {
                dprintLib("IPC operation: timeout expired.\n");
                error = DWC_TIMED;
                break;
            }
        }
#endif /* FreeBSD */
#ifdef Solaris_DW        
        error = ipc->recv_func( ipc->handle.fd, (DWC_pointer_t)((char*)buf + recvbytes), len_ptr, 0 );
#else /* !Solaris_DW */
        error = ipc->recv_func( ipc->handle, (DWC_pointer_t)((char*)buf + recvbytes), len_ptr, 0 );
#endif /* Solaris_DW */            
        if( error != DWC_NOERROR )
        {
            dprintLib( "IPC receiving: error.\n" );
            if( !ipc->nonblocking && ExpiredTimeout( tid ) )
            {
                dprintLib("IPC operation: timeout expired.\n");
                error = DWC_TIMED;
            }
            break;
        }
        recvbytes += (*len_ptr);
        if( recvbytes >= needbytes )
        {
            break;
        }
        if( ipc->nonblocking )
        {
            how = DWC_IPC_WAIT_RECV;
#ifdef Solaris_DW        
            error = ipc->wait_func( ipc->handle.fd, &how, timeout );
#else /* !Solaris_DW */
            error = ipc->wait_func( ipc->handle, &how, timeout );
#endif /* Solaris_DW */            
            if( (error != DWC_NOERROR) || !(how & DWC_IPC_WAIT_RECV) )
            {
                if( error == DWC_TIMED )
                {
                    dprintLib( "Nonblocking IPC: timeout expired.\n" );
                }
                else
                {
                    dprintLib( "Nonblocking IPC: waiting error.\n" );
                }
                break;
            }
        }
        else
        {
            if( ExpiredTimeout( tid ) )
            {
                dprintLib("IPC operation: timeout expired.\n");
                error = DWC_TIMED;
                break;
            }
        }
        *len_ptr = needbytes - recvbytes;
    }
    if( (recvbytes != needbytes) && (error == DWC_NOERROR) )
    {
        error = DWC_IPCSYSTEM;
    }
    *len_ptr = recvbytes;
    if( !ipc->nonblocking )
    {
        ClearTimeout( tid );
#if defined(FreeBSD_DW) && defined(__THREAD_BASED_DW)
        close( ipc->closer[0] );
        if( ipc->closer[1] >= 0 )
            close( ipc->closer[1] );
#endif /* FreeBSD */
    }
    return error;
}

#endif /* _DWC_IPC_C */
