#ifndef __DWC_TCP_C
#define __DWC_TCP_C

#ifndef lint
static const char rcsid[] = "$Id: dwc_tcp.c,v 1.28 2001/10/15 11:04:12 cvser Exp $";
#endif

#include <string.h>

#if defined(__UNIX_DW)
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <netdb.h>
#include <errno.h>
#if defined(EMX_DW)
#include <io.h>
#endif /* EMX_DW */
#endif /* __UNIX_DW */

#include "dwc_mutex.h"
#include "dwc_err.h"
#include "dwc_tcp.h"
#include "dwc_macro.h"
#include "dwc_log.h"

#if defined(Solaris_DW)
#ifndef INADDR_NONE
#define INADDR_NONE  -1
#endif /* INADDR_NONE */
#endif /* Solaris */

#if defined(__THREAD_BASED_DW)
DWC_mutex_t   resolve_mutex;
#endif /* __THREAD_BASED_DW */

#if defined(__WIN32_DW)
static const DWC_msg_entry_t  WinsockErrMsgTable[] = {
  { WSABASEERR,         "[0] No Error" },
  { WSAEINTR,           "[10004] Interrupted system call" },
  { WSAEBADF,           "[10009] Bad file number" },
  { WSAEACCES,          "[10013] Permission denied" },
  { WSAEFAULT,          "[10014] Bad address" },
  { WSAEINVAL,          "[10022] Invalid argument" },
  { WSAEMFILE,          "[10024] Too many open files" },
  { WSAEWOULDBLOCK,     "[10035] Operation would block" },
  { WSAEINPROGRESS,     "[10036] Operation now in progress" },
  { WSAEALREADY,        "[10037] Operation already in progress" },
  { WSAENOTSOCK,        "[10038] Socket operation on non-socket" },
  { WSAEDESTADDRREQ,    "[10039] Destination address required" },
  { WSAEMSGSIZE,        "[10040] Message too long" },
  { WSAEPROTOTYPE,      "[10041] Protocol wrong type for socket" },
  { WSAENOPROTOOPT,     "[10042] Bad protocol option" },
  { WSAEPROTONOSUPPORT, "[10043] Protocol not supported" },
  { WSAESOCKTNOSUPPORT, "[10044] Socket type not supported" },
  { WSAEOPNOTSUPP,      "[10045] Operation not supported on socket" },
  { WSAEPFNOSUPPORT,    "[10046] Protocol family not supported" },
  { WSAEAFNOSUPPORT,    "[10047] Address family not supported by protocol family" },
  { WSAEADDRINUSE,      "[10048] Address already in use" },
  { WSAEADDRNOTAVAIL,   "[10049] Can not assign requested address" },
  { WSAENETDOWN,        "[10050] Network is down" },
  { WSAENETUNREACH,     "[10051] Network is unreachable" },
  { WSAENETRESET,       "[10052] Net dropped connection or reset" },
  { WSAECONNABORTED,    "[10053] Software caused connection abort" },
  { WSAECONNRESET,      "[10054] Connection reset by peer" },
  { WSAENOBUFS,         "[10055] No buffer space available" },
  { WSAEISCONN,         "[10056] Socket is already connected" },
  { WSAENOTCONN,        "[10057] Socket is not connected" },
  { WSAESHUTDOWN,       "[10058] Can not send after socket shutdown" },
  { WSAETOOMANYREFS,    "[10059] Too many references, can't splice" },
  { WSAETIMEDOUT,       "[10060] Connection timed out" },
  { WSAECONNREFUSED,    "[10061] Connection refused" },
  { WSAELOOP,           "[10062] Too many levels of symbolic links" },
  { WSAENAMETOOLONG,    "[10063] File name too long" },
  { WSAEHOSTDOWN,       "[10064] Host is down" },
  { WSAEHOSTUNREACH,    "[10065] No Route to Host" },
  { WSAENOTEMPTY,       "[10066] Directory not empty" },
  { WSAEPROCLIM,        "[10067] Too many processes" },
  { WSAEUSERS,          "[10068] Too many users" },
  { WSAEDQUOT,          "[10069] Disc Quota Exceeded" },
  { WSAESTALE,          "[10070] Stale NFS file handle" },
  { WSAEREMOTE,         "[10071] Too many levels of remote in path" },
  { WSASYSNOTREADY,     "[10091] Network SubSystem is unavailable" },
  { WSAVERNOTSUPPORTED, "[10092] WINSOCK DLL Version out of range" },
  { WSANOTINITIALISED,  "[10093] Successful WSASTARTUP not yet performed" },
  { WSAHOST_NOT_FOUND,  "[11001] Host not found" },
  { WSATRY_AGAIN,       "[11002] Non-Authoritative Host not found" },
  { WSANO_RECOVERY,     "[11003] Non-Recoverable errors: FORMERR, REFUSED, NOTIMP" },
  { WSANO_DATA,         "[11004] Valid name, no data record of requested" },
  END_OF_TABLE
};
static const int  WinsockErrMsgTableSize = (sizeof(WinsockErrMsgTable) / sizeof(DWC_msg_entry_t)) - 1;
#endif /* __WIN32_DW */

/* -------------------------------------------------------------------- */
static DWC_string_t tcpsys_error ( void )
{
#if defined(__WIN32_DW)
    return findWhat( WSAGetLastError(), WinsockErrMsgTable );
#endif /* __WIN32_DW */
#if defined(__UNIX_DW)
    return stdlib_error( errno );
#endif /* __UNIX_DW */
}

/* -------------------------------------------------------------------- */
DWC_error_code_t init_tcp_sockets ( void )
{
#if defined(__WIN32_DW)
    WORD    wVersionRequested;
    WSADATA wsaData;

    wVersionRequested = MAKEWORD( 1, 1 );
    if( WSAStartup( wVersionRequested, &wsaData ) != 0 )
    {
        dprintLib("TCP sockets: can not inialize WinSock subsystem - %s.\n", tcpsys_error() );
        return DWC_IPCSYSTEM;
    }
    if( (LOBYTE( wsaData.wVersion ) != 1)
        ||
        (HIBYTE( wsaData.wVersion ) != 1) )
    {
        dprintLib("TCP sockets: can not find WinSock 1.1.\n");
        return DWC_IPCSYSTEM;
    }
#endif /* __WIN32_DW */
#if defined(__THREAD_BASED_DW)
    resolve_mutex = dwc_mutex_create( "ResolverMutex" );
    if( !resolve_mutex )
    {
#if defined(__WIN32_DW)
        WSACleanup();
#endif /* __WIN32_DW */
        return DWC_NORESOURCE;
    }
#endif /* __THREAD_BASED_DW */
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t dinit_tcp_sockets ( void )
{
#if defined(__THREAD_BASED_DW)
    DWC_bool_t   error = FALSE;

    if( !dwc_mutex_lock( resolve_mutex ) )
    {
        error = TRUE;
    }
    else
    {
        if( !dwc_mutex_unlock( resolve_mutex ) )
        {
            error = TRUE;
        }
    }
    if( dwc_mutex_destroy( resolve_mutex ) != DWC_NOERROR )
    {
        error = TRUE;
    }
    if( error  )
    {
        dprintLib( "TCP sockets: some errors during deinit tcp system.\n" );
        return DWC_IPCSYSTEM;
    }
#endif /* __THREAD_BASED_DW */
#if defined(__WIN32_DW)
    if( WSACleanup() != 0 )
    {
        dprintLib("TCP sockets: can not deinitialize WinSock subsystem - %s.\n", tcpsys_error() );
        return DWC_IPCSYSTEM;
    }
#endif /* __WIN32_DW */
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t resolve_tcp_addr( DWC_tcp_socket_addr_t* naddr, DWC_string_t name )
{
    struct hostent      *he = NULL;
    DWC_error_code_t     error;

    if( !name || (name[0] == NULLCHAR) )
    {
        dprintLib( "TCP sockets: empty name passed to resolver.\n" );
        return DWC_BADPARAMS;
    }
#if defined(__WIN32_DW) || defined(Solaris_DW) || defined(EMX_DW)
    naddr->sin_addr.s_addr = inet_addr( name );
    if( naddr->sin_addr.s_addr == INADDR_NONE )
    {
#else /* !(__WIN32_DW || Solaris || EMX) */
    if( inet_aton( name, &naddr->sin_addr ) == 0 )
    {
#endif /* __WIN32_DW || Solaris */
#if defined(__THREAD_BASED_DW)
        if( !dwc_mutex_lock( resolve_mutex ) )
        {
            return DWC_IPCSYSTEM;
        }
#endif /* __THREAD_BASED_DW */
        dprintLib("TCP sockets: call resolver for %s\n", name);
        he = gethostbyname( name );
        if( he )
        {
            if( (he->h_addrtype == AF_INET)
                && (he->h_addr_list != NULL)
                && (he->h_addr_list[0] != NULL)
                && (he->h_length == sizeof(naddr->sin_addr)) )
            {
                memcpy( &naddr->sin_addr, he->h_addr_list[0], sizeof(naddr->sin_addr) );
                dprintLib("TCP sockets: resolver return %s\n", inet_ntoa( naddr->sin_addr ) );
                error = DWC_NOERROR;
            }
            else
            {
                dprintLib("TCP sockets: resolver return corrupted address.\n");
                error = DWC_INVALIDADDR;
            }
        }
        else
        {
#if defined(__WIN32_DW) || defined(Solaris_DW) || defined(EMX_DW)
            dprintLib("TCP sockets: gethostbyname() error : %s.\n", tcpsys_error() );
#else /* !(__WIN32_DW || Solaris || EMX) */
            dprintLib("TCP sockets: gethostbyname() error : %s.\n", hstrerror(h_errno) );
#endif /*__WIN32_DW || Solaris */
            error = DWC_INVALIDADDR;
        }
#if defined(__THREAD_BASED_DW)
        if( !dwc_mutex_unlock( resolve_mutex ) )
        {
            return DWC_IPCSYSTEM;
        }
#endif /* __THREAD_BASED_DW */
    }
    else
    {
        error = DWC_NOERROR;
    }
    return error;
}


/* -------------------------------------------------------------------- */
DWC_error_code_t addr_tcp_op ( DWC_tcp_socket_addr_t* addr, DWC_cmd_t cmd, DWC_string_t name, va_list args )
{
    unsigned short       port;
    DWC_bool_t           used = FALSE;
    DWC_error_code_t     error;

    if( (cmd == DWC_IPC_CMD_SETADDR)
        ||
        (cmd == DWC_IPC_CMD_SETALL) )
    {
        used = TRUE;
        addr->sin_addr = va_arg( args, struct in_addr );
        if( addr->sin_addr.s_addr == INADDR_NONE )
        {
            error = DWC_BADPARAMS;
            goto error_exit;
        }
    }
    if( (cmd == DWC_IPC_CMD_SETPORT)
        ||
        (cmd == DWC_IPC_CMD_SETALL) )
    {
        used = TRUE;
        port = (unsigned short)(0xFFFF & va_arg( args, unsigned int ));
        if( port == 0 )
        {
            error = DWC_BADPARAMS;
            goto error_exit;
        }
        addr->sin_port = htons( port );
    }
    if( cmd == DWC_IPC_CMD_RESOLVE )
    {
        used = TRUE;
        error = resolve_tcp_addr( addr, name );
        if( error != DWC_NOERROR )
        {
            goto error_exit;
        }
    }
    if( !used )
    {
        error = DWC_BADPARAMS;
    }
    else
    {
        error = DWC_NOERROR;
    }
  error_exit:
    return error;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t create_tcp_socket ( DWC_tcp_socket_t *sock, DWC_bool_t nonblocking )
{
#if defined(__WIN32_DW)
    int         err;
#endif /* __WIN32_DW */
#if defined(__UNIX_DW)
    int         opt;
#endif /* __UNIX_DW */

    PARAM_ASSERT( sock );

    *sock = socket( PF_INET, SOCK_STREAM, IPPROTO_TCP );
    if( (*sock) == INVALID_TCP_SOCKET )
    {
        dprintLib("TCP sockets: creation failed - %s.\n", tcpsys_error() );
        return DWC_IPCSYSTEM;
    }
    if( nonblocking )
    {
#if defined(__UNIX_DW)
        opt = fcntl( (*sock), F_GETFL, 0 );
        if( fcntl( (*sock), F_SETFL, opt | O_NONBLOCK ) == -1 )
        {
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
        err = ioctlsocket( (*sock), FIONBIO, &nonblocking );
        if( SOCK_ERR( err ) )
        {
#endif /* __WIN32_DW */
            dprintLib("TCP sockets: switching to nonblocking mode failed - %s.\n", tcpsys_error() );
            close_tcp_socket( sock );
            return DWC_IPCSYSTEM;
        }
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t bind_tcp_socket ( DWC_tcp_socket_t sock, DWC_tcp_socket_addr_t* addr )
{
    DWC_socklen_t len = sizeof(DWC_tcp_socket_addr_t);

    if( bind( sock, (struct sockaddr*) addr, len ) < 0 )
    {
        dprintLib("TCP sockets: bind() failed - %s.\n", tcpsys_error() );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t listen_tcp_socket ( DWC_tcp_socket_t sock, int backlog )
{
    if( listen( sock, backlog ) != 0 )
    {
        dprintLib("TCP sockets: listen() failed - %s.\n", tcpsys_error() );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t accept_tcp_socket ( DWC_tcp_socket_t sock, DWC_tcp_socket_addr_t *addr, DWC_tcp_socket_t *newsock )
{
    DWC_socklen_t len = sizeof(DWC_tcp_socket_addr_t);

    *newsock = accept( sock, (struct sockaddr*) addr, &len );
    if( (*newsock) == INVALID_TCP_SOCKET )
    {
        dprintLib("TCP sockets: accept() failed - %s.\n", tcpsys_error() );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t connect_tcp_socket ( DWC_tcp_socket_t sock, DWC_tcp_socket_addr_t *addr, DWC_timeout_t timeout )
{
    DWC_error_code_t  err;
    int               how;

    dprintLib("TCP sockets: connecting to %s:%d.\n", inet_ntoa( addr->sin_addr ), ntohs(addr->sin_port) );
    if( connect( sock, (struct sockaddr*) addr, sizeof(DWC_tcp_socket_addr_t) ) < 0 )
    {
        if( IN_PROGRESS )
        {
            how = DWC_IPC_WAIT_EXCP | DWC_IPC_WAIT_SEND;
            err = wait_tcp_socket( sock, &how, timeout );
            if( err == DWC_NOERROR  )
            {
                if( how & DWC_IPC_WAIT_SEND )
                {
                    return DWC_NOERROR;
                }
                err = DWC_IPCSYSTEM;
                dprintLib("TCP sockets: connect() failed.\n");
            }
            else if( err == DWC_TIMED )
            {
                dprintLib("TCP sockets: connect() failed - timeout.\n");
            }
            else
            {
                /* asv - add diagnostic */
                dprintLib("TCP sockets: connect() failed.\n");
            }
            return err;
        }
        dprintLib("TCP sockets: connect() failed - %s.\n", tcpsys_error() );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t send_tcp_socket ( DWC_tcp_socket_t sock, DWC_pointer_t buf, DWC_length_t *len, int options )
{
    int sent;

    sent = send( sock, buf, *len, 0 );
    if( SOCK_ERR(sent) )
    {
        (*len) = 0;
        if( WOULD_BLOCK )
        {
            return DWC_NOERROR;
        }
        dprintLib("TCP sockets: send() failed - %s.\n", tcpsys_error() );
        return DWC_IPCSYSTEM;
    }
    (*len) = sent;
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t recv_tcp_socket ( DWC_tcp_socket_t sock, DWC_pointer_t buf, DWC_length_t *len, int options )
{
    int readed;

    readed = recv( sock, buf, *len, options );
    if( SOCK_ERR(readed) )
    {
        (*len) = 0;
        if( WOULD_BLOCK )
        {
            return DWC_NOERROR;
        }
        dprintLib( "TCP sockets: recv() failed - %s.\n", tcpsys_error() );
        return DWC_IPCSYSTEM;
    }
    if( readed == 0 )
    {
        dprintLib( "TCP sockets: recv() req=%d failed - connection closed by remote side.\n", *len );
        (*len) = 0;
        return DWC_IPCSYSTEM;
    }
    if( readed != (*len) )
    {
        dprintLib( "TCP sockets: recv() req=%d, readed=%d.\n", *len, readed );
    }
    (*len) = readed;
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t wait_tcp_socket( DWC_tcp_socket_t sock, int *how, DWC_timeout_t timeout )
{
    fd_set         rd_set, wr_set, ex_set;
    struct timeval timo, *pt;
    int            rd;

    FD_ZERO( &rd_set );
    FD_ZERO( &wr_set );
    FD_ZERO( &ex_set );
    if( (*how & DWC_IPC_WAIT_RECV) == DWC_IPC_WAIT_RECV )
    {
        FD_SET( sock, &rd_set );
    }
    if( (*how & DWC_IPC_WAIT_SEND) == DWC_IPC_WAIT_SEND )
    {
        FD_SET( sock, &wr_set );
    }
    if( (*how & DWC_IPC_WAIT_EXCP) == DWC_IPC_WAIT_EXCP )
    {
        FD_SET( sock, &ex_set );
    }
    *how = 0;
    pt = NULL;
    if( timeout > 0 )
    {
        timo.tv_sec = timeout / 1000000;
        timo.tv_usec = timeout % 1000000;
        pt = &timo;
    }
    rd = select( sock+1, &rd_set, &wr_set, &ex_set, pt );
    if( (rd == 0) && pt )
    {
        return DWC_TIMED;
    }
    if( rd < 0 )
    {
        dprintLib("TCP sockets: select() failed - %s.\n", tcpsys_error() );
        return DWC_IPCSYSTEM;
    }
    if( FD_ISSET( sock, &rd_set ) )
    {
        *how |= DWC_IPC_WAIT_RECV;
    }
    if( FD_ISSET( sock, &wr_set ) )
    {
        *how |= DWC_IPC_WAIT_SEND;
    }
    if( FD_ISSET( sock, &ex_set ) )
    {
        *how |= DWC_IPC_WAIT_EXCP;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t shutdown_tcp_socket ( DWC_tcp_socket_t *sock )
{
    int error;

    error = shutdown( *sock, SHUT_RDWR );
    (*sock) = INVALID_TCP_SOCKET;
    if( error )
    {
        dprintLib("TCP sockets: shutdown() failed - %s.\n", tcpsys_error() );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t close_tcp_socket ( DWC_tcp_socket_t *sock )
{
    int error;

#if defined(__WIN32_DW)
    error = closesocket( *sock );
#else
    error = close( *sock );
#endif
    *sock = INVALID_TCP_SOCKET;
    if( error )
    {
        dprintLib("TCP sockets: close() failed - %s.\n", tcpsys_error() );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------- */
void cancel_tcp_call( DWC_pointer_t sock )
{
    dprintLib("TCP sockets: cancelling of syscall called - timeout.\n");
#if defined(FreeBSD_DW) && defined(__THREAD_BASED_DW)
    if( close_tcp_socket( (DWC_tcp_socket_t*)(sock) ) != DWC_NOERROR )
#else /* ! (FreeBSD_DW && __THREAD_BASED_DW) */
    if( shutdown_tcp_socket( (DWC_tcp_socket_t*)(sock) ) != DWC_NOERROR )
#endif /* FreeBSD */
    {
        dprintLib("TCP sockets: cancelling of blocking syscall failed.\n" );
    }
}

#endif /* __DWC_TCP_H */
