#ifndef __DWC_TIMO_C
#define __DWC_TIMO_C

#ifndef lint
static const char rcsid[] = "$Id: dwc_timo.c,v 1.27 2001/10/08 15:48:52 cvser Exp $";
#endif

#if defined(__SIGNAL_BASED_DW)
#include <sys/time.h>
#include <signal.h>
#endif /* __SIGNAL_BASED_DW */
#include <errno.h>
#include <string.h>

#include "dwc_timo.h"
#include "dwc_lib.h"
#include "dwc_type.h"
#include "dwc_err.h"
#include "dwc_mutex.h"
#include "dwc_event.h"
#include "dwc_thread.h"
#include "dwc_macro.h"
#include "bsd_queue.h"

typedef struct _DWC_timer_t {
    TAILQ_ENTRY( _DWC_timer_t )     all;
    TAILQ_ENTRY( _DWC_timer_t )     active;
    DWC_bool_t                      alarmed;
    DWC_bool_t                      waited;
    DWC_bool_t                      deletable;
    DWC_timeout_t                   value;
    DWC_TimeoutProc_t               proc;
    DWC_pointer_t                   data;
} DWC_timer_t;

typedef TAILQ_HEAD( _DWC_tqhead_t, _DWC_timer_t ) DWC_tqhead_t;

typedef struct _DWC_tqueue_t {
    DWC_tqhead_t                    all_head;
    DWC_tqhead_t                    active_head;
    DWC_bool_t                      inited;
    DWC_bool_t                      locked;
    DWC_timeout_t                   slice;
#if defined(__THREAD_BASED_DW)
#define EVENTS_NUM           (3)
#define EV_NEW_VALUE         (0)
#define EV_NOW_EXIT          (1)
#define EV_TIMER_STOPED      (2)
    DWC_mutex_t                     mutex;
    DWC_event_t                     events[EVENTS_NUM];
    DWC_thread_id_t                 locker;
#endif /* __THREAD_BASED_DW */
#if defined(__SIGNAL_BASED_DW)
    struct sigaction                oldact;
    sigset_t                        sigset;
    DWC_bool_t                      in_signal;
#endif /* __SIGNAL_BASED_DW */
} DWC_tqueue_t;

static DWC_tqueue_t      TimerQueue = {
    /* all_head */    TAILQ_HEAD_INITIALIZER( TimerQueue.all_head ),
    /* active head */ TAILQ_HEAD_INITIALIZER( TimerQueue.active_head ),
    /* inited */      FALSE,
    /* locked */      FALSE,
    /* slice */       INFINITE_TIMEOUT
    /* others components are compile-depend */
};

static DWC_bool_t           SelfLocking( void );
static DWC_bool_t           LockQueue ( DWC_tqueue_t *queue );
static DWC_bool_t           UnLockQueue ( DWC_tqueue_t *queue );
static DWC_error_code_t     RunIntTimer( DWC_timeout_t value );
static DWC_error_code_t     StopIntTimer( void );
static void                 ReceiveTick( void );
static void                 AlarmTimer( DWC_timer_hnd_t Timer );
static void                 RemoveTimer( DWC_timer_hnd_t Timer, DWC_bool_t Finally );

#if defined(__UNIX_DW)
#if defined(__SIGNAL_BASED_DW)
#ifndef EMX_DW
#define TIMER_NUM_SLICES    (16)
#else /* EMX */
#define TIMER_NUM_SLICES    (2)
#endif /* EMX */
#endif /* __SIGNAL_BASED_DW */
#if defined(__THREAD_BASED_DW)
#define TIMER_NUM_SLICES    (2)
#endif /* __THREAD_BASED_DW */
#if defined(Linux_DW)
#define TIMER_MIN_SLICE     (10)
#elif defined(EMX_DW)
#define TIMER_MIN_SLICE     (1000)
#else /* FreeBSD. Default for other - not tested */
#define TIMER_MIN_SLICE     (20)
#endif /* Linux */
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
#define TIMER_NUM_SLICES    (2)
#define TIMER_MIN_SLICE     (55) /* 55 msec for Win95 better */
#endif /* __WIN32_DW */

#if defined(__SIGNAL_BASED_DW)
/* --------------------------------------------------------------------- */
static void UnixSigAlarm ( int signo )
{
    if( signo == SIGALRM )
    {
        TimerQueue.in_signal = TRUE;
        ReceiveTick();
        TimerQueue.in_signal = FALSE;
    }
}
#endif /* __SIGNAL_BASED_DW */

#if defined(__THREAD_BASED_DW)
/* --------------------------------------------------------------------- */
static void TimerThread ( DWC_thread_arg_t arg )
{
    DWC_bool_t      need_exit = FALSE;
    int             retcode;

    while( !need_exit )
    {
        retcode = dwc_event_multiple_wait( TimerQueue.events, EVENTS_NUM, TimerQueue.slice );
        if( retcode == EVENTS_NUM )
        {
            /* timeouted */
            ReceiveTick();
        }
        else if( retcode == EV_NEW_VALUE )
        {
            /* re-setup timeout */
            continue;
        }
        else if( retcode == EV_NOW_EXIT )
        {
            /* ExitThread event */
            need_exit = TRUE;
            dprintLib( "timer_thread: should be stopped.\n", retcode );
        }
        else
        {
            /* any error */
            need_exit = TRUE;
            dprintLib( "timer_thread: (%d) errors while events waiting.\n", retcode );
        }
    }
    if( !dwc_event_set( TimerQueue.events[EV_TIMER_STOPED] ) )
    {
        dprintLib( "timer_engine: timer-thread can not report about stopping.\n" );
    }
}
#endif /* __THREAD_BASED_DW */

/* --------------------------------------------------------------------- */
DWC_error_code_t InitTimerEngine( void )
{
#if defined(__SIGNAL_BASED_DW)
    struct sigaction  sig;
#endif /* __SIGNAL_BASED_DW */

    if( TimerQueue.inited )
    {
        dprintLib( "Timer engine is already initalized.\n" );
        return DWC_ALREADYPERFORM;
    }
    TimerQueue.locked = TRUE;
    TimerQueue.slice = 0;
#if defined(__SIGNAL_BASED_DW)
    if( StopIntTimer() != DWC_NOERROR )
    {
        dprintLib( "Can not set timer - [%u] %s.\n", errno, strerror(errno) );
        return DWC_NORESOURCE;
    }
    sig.sa_handler = UnixSigAlarm;
#ifndef EMX_DW
    sig.sa_flags = SA_RESTART;
#endif /* EMX */
    sigemptyset( &sig.sa_mask );
    if( sigaction( SIGALRM, &sig, &TimerQueue.oldact ) < 0 )
    {
        dprintLib( "Can not set signal handler for SIGALRM - [%u] %s.\n", errno, strerror(errno) );
        return DWC_COMMONERROR;
    }
    sigemptyset( &TimerQueue.sigset );
    sigaddset( &TimerQueue.sigset, SIGALRM );
#endif /* __SIGNAL_BASED_DW */
#if defined(__THREAD_BASED_DW)
    TimerQueue.mutex = dwc_mutex_create( "TimerQueueLock" );
    if( !TimerQueue.mutex )
    {
        return DWC_NORESOURCE;
    }
    TimerQueue.events[EV_NEW_VALUE] = dwc_event_create( "NewValueForTimer" );
    if( !TimerQueue.events[EV_NEW_VALUE] )
    {
        dwc_mutex_destroy( TimerQueue.mutex );
        return DWC_NORESOURCE;
    }
    TimerQueue.events[EV_NOW_EXIT] = dwc_event_create( "StopTimerThread" );
    if( !TimerQueue.events[EV_NOW_EXIT] )
    {
        dwc_event_destroy( TimerQueue.events[EV_NEW_VALUE] );
        dwc_mutex_destroy( TimerQueue.mutex );
        return DWC_NORESOURCE;
    }
    TimerQueue.events[EV_TIMER_STOPED] = dwc_event_create( "TimerThreadExited" );
    if( !TimerQueue.events[EV_TIMER_STOPED] )
    {
        dwc_event_destroy( TimerQueue.events[EV_NEW_VALUE] );
        dwc_event_destroy( TimerQueue.events[EV_NOW_EXIT] );
        dwc_mutex_destroy( TimerQueue.mutex );
        return DWC_NORESOURCE;
    }
    if( dwc_thread_create( TimerThread, NULL ) != DWC_NOERROR )
    {
        dwc_event_destroy( TimerQueue.events[EV_NEW_VALUE] );
        dwc_event_destroy( TimerQueue.events[EV_NOW_EXIT] );
        dwc_event_destroy( TimerQueue.events[EV_TIMER_STOPED] );
        dwc_mutex_destroy( TimerQueue.mutex );
        return DWC_NORESOURCE;
    }
#endif /* __THREAD_BASED_DW */
    TimerQueue.locked = FALSE;
    TimerQueue.inited = TRUE;
    return DWC_NOERROR;
}

/* --------------------------------------------------------------------- */
DWC_error_code_t DeInitTimerEngine( void )
{
#if defined(__THREAD_BASED_DW)
    DWC_error_code_t  retcode;
#endif /* __THREAD_BASED_DW */
    DWC_bool_t        error = FALSE;

    if( !TimerQueue.inited )
    {
        dprintLib( "Timer engine is not initalized.\n" );
        return DWC_NOTREADY;
    }
    if( SelfLocking() )
    {
        dprintLib( "Timer engine locked: you shouldn not call Timeout-routines from your timeout handler.\n" );
        return DWC_NOTREADY;
    }
    if( !LockQueue( &TimerQueue ) )
    {
        dprintLib( "Can not lock engine.\n" );
        error = TRUE;
    }
    else
    {
        while( !TAILQ_EMPTY( &TimerQueue.all_head ) )
        {
            RemoveTimer( TAILQ_FIRST( &TimerQueue.all_head), TRUE );
        }
        if( !UnLockQueue( &TimerQueue ) )
        {
            dprintLib( "Can not unlock engine.\n" );
            error = TRUE;
        }
    }
#if defined(__SIGNAL_BASED_DW)
    if( StopIntTimer() != DWC_NOERROR )
    {
        dprintLib( "Can not stop timer.\n" );
        error = TRUE;
    }
    if( sigaction( SIGALRM, &TimerQueue.oldact, NULL ) < 0 )
    {
        dprintLib( "Can not restore old action for SIGALRM - [%u] %s.\n", errno, strerror(errno) );
        error = TRUE;
    }
#endif /* __SIGNAL_BASED_DW */
#if defined(__THREAD_BASED_DW)
    if( dwc_event_set( TimerQueue.events[EV_NOW_EXIT] ) )
    {
        retcode = dwc_event_wait( TimerQueue.events[EV_TIMER_STOPED], 10000 );
        if( retcode == DWC_TIMED )
        {
            dprintLib( "deinit_timer_engine: waiting for timer stop timed out.\n" );
            error = TRUE;
        }
        else if( retcode != DWC_NOERROR )
        {
            dprintLib( "deinit_timer_engine: waiting for timer stop failed.\n" );
            error = TRUE;
        }
    }
    else
    {
        dprintLib( "deinit_timer_engine: setting event for timer failed.\n" );
        error = TRUE;
    }
    if( dwc_event_destroy( TimerQueue.events[EV_NEW_VALUE] ) != DWC_NOERROR )
    {
        error = TRUE;
    }
    if( dwc_event_destroy( TimerQueue.events[EV_NOW_EXIT] ) != DWC_NOERROR )
    {
        error = TRUE;
    }
    if( dwc_event_destroy( TimerQueue.events[EV_TIMER_STOPED] ) != DWC_NOERROR )
    {
        error = TRUE;
    }
    if( dwc_mutex_destroy( TimerQueue.mutex ) != DWC_NOERROR )
    {
        error = TRUE;
    }
#endif /* __THREAD_BASED_DW */
    return ( !error ? DWC_NOERROR : DWC_COMMONERROR );
}

/* --------------------------------------------------------------------- */
static DWC_bool_t LockQueue ( DWC_tqueue_t *queue )
{
#if defined(__SIGNAL_BASED_DW)
    if( !queue->in_signal )
    {
        if( sigprocmask( SIG_BLOCK, &TimerQueue.sigset, NULL ) < 0 )
        {
            dprintLib(" Can not block signals - [%u] %s.\n", errno, strerror(errno) );
            return FALSE;
        }
        queue->locked = TRUE;
    }
    return TRUE;
#endif /* __SIGNAL_BASED_DW */
#if defined(__THREAD_BASED_DW)
    if( dwc_mutex_lock( queue->mutex ) )
    {
        dwc_thread_get_id( &queue->locker );
        queue->locked = TRUE;
        return TRUE;
    }
    else
    {
        return FALSE;
    }
#endif /* __THREAD_BASED_DW */
}

/* --------------------------------------------------------------------- */
static DWC_bool_t UnLockQueue ( DWC_tqueue_t *queue )
{
#if defined(__SIGNAL_BASED_DW)
    if( !queue->in_signal )
    {
        queue->locked = FALSE;
        if( sigprocmask( SIG_UNBLOCK, &TimerQueue.sigset, NULL ) < 0 )
        {
            dprintLib(" Can not unblock signals - [%u] %s.\n", errno, strerror(errno) );
            return FALSE;
        }
    }
    return TRUE;
#endif /* __SIGNAL_BASED_DW */
#if defined(__THREAD_BASED_DW)
    queue->locked = FALSE;
    return dwc_mutex_unlock( queue->mutex );
#endif /* __THREAD_BASED_DW */
}

/* --------------------------------------------------------------------- */
static DWC_error_code_t RunIntTimer( DWC_timeout_t timeout )
{
#if defined(__SIGNAL_BASED_DW) && !defined(EMX_DW)
    struct itimerval  itv;
#endif /* __SIGNAL_BASED_DW && ! EMX */

    if( timeout == INFINITE_TIMEOUT )
    {
        TimerQueue.slice = INFINITE_TIMEOUT;
    }
    else if( timeout < TIMER_MIN_SLICE * TIMER_NUM_SLICES )
    {
        TimerQueue.slice = TIMER_MIN_SLICE;
    }
    else
    {
        TimerQueue.slice = timeout / TIMER_NUM_SLICES;
        /* all systems has fixed quantized step of timers */
        TimerQueue.slice -= TimerQueue.slice % TIMER_MIN_SLICE;
    }
#if defined(__SIGNAL_BASED_DW)
#ifndef EMX_DW
    itv.it_value.tv_sec = TimerQueue.slice / 1000;
    itv.it_value.tv_usec = (TimerQueue.slice % 1000) * 1000;
    itv.it_interval.tv_sec = itv.it_value.tv_sec;
    itv.it_interval.tv_usec = itv.it_value.tv_usec;
    if( setitimer( ITIMER_REAL, &itv, NULL ) < 0 )
    {
        dprintLib("setitimer() error - [%u] %s.\n", errno, strerror(errno) );
        return DWC_COMMONERROR;
    }
#else /* EMX */
    /* OS/2 EMX doesnot have setitimer */
	  alarm( TimerQueue.slice / 1000 );
#endif /* EMX */
#endif /* __SIGNAL_BASED_DW */
#if defined(__THREAD_BASED_DW)
    if( !dwc_event_set( TimerQueue.events[EV_NEW_VALUE] ) )
    {
        return DWC_COMMONERROR;
    }
#endif /* __THREAD_BASED_DW */
    return DWC_NOERROR;
}

/* --------------------------------------------------------------------- */
static DWC_error_code_t StopIntTimer( void )
{
    return RunIntTimer( INFINITE_TIMEOUT );
}

/* --------------------------------------------------------------------- */
static DWC_bool_t SelfLocking( void )
{
#if defined(__SIGNAL_BASED_DW)
    return TimerQueue.in_signal;
#endif /* __SIGNAL_BASED_DW */
#if defined(__THREAD_BASED_DW)
   /* asv: it looks unsafe, but it should work, because
 if queue
      was locked before next check to equal
 it can be unlocked
      and locked again, BUT not this thread (we
 here) end checks
      fails. If we have mutex (and set locked),
 noboby can't unlock
      and between two checks state of locked
 can't be changed */
    if( TimerQueue.locked && dwc_thread_isself( &TimerQueue.locker )  )
    {
        dprintLib( "timer_engine: possible selflocking -> prevent.\n" );
        return TRUE;
    }
    else
    {
        return FALSE;
    }
#endif /* __THREAD_BASED_DW */
}

/* --------------------------------------------------------------------- */
static void ReceiveTick( void )
{
    DWC_timer_hnd_t   Timer;
    DWC_timeout_t     timo;

    if( !LockQueue( &TimerQueue ) )
    {
        dprintLib( "Info: timer handler can not lock engine - lose slice.\n" );
        return;
    }
    timo = TimerQueue.slice;
  cont_l:
    Timer = TAILQ_FIRST( &TimerQueue.active_head );
    if( !Timer )
    {
        StopIntTimer();
    }
    else if( timo < Timer->value )
    {
        Timer->value -= timo;
    }
    else
    {
        timo -= Timer->value;
        Timer->value = 0;
        AlarmTimer( Timer );
        goto cont_l;
    }
    if( !UnLockQueue( &TimerQueue ) )
    {
        dprintLib( "Error: can not unlock engine.\n" );
    }
}

/* --------------------------------------------------------------------- */
static void AlarmTimer( DWC_timer_hnd_t Timer )
{
    if( !Timer->alarmed )
    {
        Timer->alarmed = TRUE;
        if( Timer->proc )
        {
            Timer->proc( Timer->data );
        }
        RemoveTimer( Timer, Timer->deletable );
    }
    else
    {
        wprintLib( "timer_thread: ERROR - timer is already alarmed.\n" );
    }
}

/* --------------------------------------------------------------------- */
static void RemoveTimer( DWC_timer_hnd_t Timer, DWC_bool_t Finally )
{
    DWC_timer_hnd_t  anext;

    if( !Timer->waited )
    {
        Timer->waited = TRUE;
        TAILQ_REMOVE( &TimerQueue.active_head, Timer, active );
        if( TAILQ_EMPTY( &TimerQueue.active_head ) )
        {
            if( StopIntTimer() != DWC_NOERROR )
            {
                dprintLib( "Error: Can not stop internal timer.\n" );
            }
        }
        else
        {
            anext = TAILQ_NEXT( Timer, active );
            if( anext )
            {
                if( Timer->value > 0 )
                {
                    anext->value += Timer->value;
                    Timer->value = 0;
                }
            }
            if( RunIntTimer( TAILQ_FIRST( &TimerQueue.active_head )->value ) != DWC_NOERROR )
            {
                dprintLib( "Error: Can not set internal timer.\n" );
            }
        }
    }
    if( Finally )
    {
        TAILQ_REMOVE( &TimerQueue.all_head, Timer, all );
        DWC_Free( Timer );
    }
}

/* --------------------------------------------------------------------- */
DWC_timer_hnd_t SetTimeout( DWC_timeout_t timeout, DWC_TimeoutProc_t Proc, DWC_pointer_t Data, DWC_bool_t DelAtAlarm )
{
    DWC_timer_hnd_t   Timer;
    DWC_timer_hnd_t   tmpTimer;
    DWC_bool_t        inserted = FALSE;

    if( !TimerQueue.inited )
    {
        dprintLib( "Timer engine is not initalized.\n" );
        return INVALID_TIMER;
    }
    /* It's guard *ONLY* against one thread loop:
     *  alarm -> lock(); call -> user_proc() -> and user call us recursively lock()
     */
    if( SelfLocking() )
    {
        return INVALID_TIMER;
    }
    if( timeout != INVALID_TIMEOUT )
    {
        if( timeout == INFINITE_TIMEOUT )
        {
            return INFINITE_TIMER;
        }
        Timer = DWC_Malloc( DWC_timer_hnd_t, sizeof(DWC_timer_t) );
        if( !Timer )
        {
            dprintLib( "Can not create timer - no memory.\n" );
            return INVALID_TIMER;
        }
        if( !LockQueue( &TimerQueue ) )
        {
            dprintLib( "Can not lock timer engine on create.\n" );
            DWC_Free( Timer );
            return INVALID_TIMER;
        }
        Timer->alarmed = FALSE;
        Timer->deletable = DelAtAlarm;
        Timer->waited = FALSE;
        Timer->proc = Proc;
        Timer->data = Data;
        if( TAILQ_EMPTY( &TimerQueue.active_head ) )
        {
            TAILQ_INSERT_HEAD( &TimerQueue.active_head, Timer, active) ;
            RunIntTimer( timeout );
        }
        else if( timeout < TAILQ_FIRST(&TimerQueue.active_head)->value )
        {
            TAILQ_FIRST(&TimerQueue.active_head)->value -= timeout;
            TAILQ_INSERT_HEAD(&TimerQueue.active_head, Timer, active);
            RunIntTimer( timeout );
        }
        else
        {
            TAILQ_FOREACH( tmpTimer, &TimerQueue.active_head, active )
            {
                if( timeout < tmpTimer->value )
                {
                    TAILQ_INSERT_BEFORE( tmpTimer, Timer, active );
                    tmpTimer->value -= timeout;
                    inserted = TRUE;
                    break;
                }
                timeout -= tmpTimer->value;
            }
            if( !inserted )
            {
                TAILQ_INSERT_TAIL( &TimerQueue.active_head, Timer, active );
            }
        }
        Timer->value = timeout;
        TAILQ_INSERT_TAIL( &TimerQueue.all_head, Timer, all );
        if( !UnLockQueue( &TimerQueue ) )
        {
            dprintLib( "Error: can't unlock timer engine.\n" );
            RemoveTimer( Timer, TRUE );
            DWC_Free( Timer );
            return INVALID_TIMER;
        }
        return Timer;
    }
    else
    {
        return INVALID_TIMER;
    }
}

/* --------------------------------------------------------------------- */
DWC_timeout_t RetrieveTimeout( DWC_timer_hnd_t Timer )
{
    DWC_timeout_t     retval = 0;
    DWC_timer_hnd_t   tmpTimer;
    DWC_bool_t        found = FALSE;

    if( !TimerQueue.inited )
    {
        dprintLib( "Timer engine is not initalized.\n" );
        return INVALID_TIMEOUT;
    }
    if( Timer == INVALID_TIMER )
    {
        dprintLib( "Invalid timer requested.\n" );
        return INVALID_TIMEOUT;
    }
    else if( Timer == INFINITE_TIMER )
    {
        return INFINITE_TIMEOUT;
    }
    /* It's guard *ONLY* against one thread loop:
     *  alarm -> lock(); call -> user_proc() -> and user call us recursively lock()
     */
    if( SelfLocking() )
    {
        return INVALID_TIMEOUT;
    }
    if( !LockQueue( &TimerQueue ) )
    {
        dprintLib( "Can not lock timer engine.\n" );
        return INVALID_TIMEOUT;
    }
    TAILQ_FOREACH( tmpTimer, &TimerQueue.active_head, active )
    {
        retval += tmpTimer->value;
        if( tmpTimer == Timer )
        {
            found = TRUE;
            break;
        }
    }
    if( !UnLockQueue( &TimerQueue ) )
    {
        dprintLib( "Can not unlock timer engine.\n" );
        return INVALID_TIMEOUT;
    }
    if( found )
    {
        return retval;
    }
    else
    {
        dprintLib( "Invalid timer: not existing or already alarmed.\n" );
        return INVALID_TIMEOUT;
    }
}

/* --------------------------------------------------------------------- */
DWC_bool_t AlarmTimeout( DWC_timer_hnd_t Timer )
{
    DWC_timer_hnd_t   tmpTimer;
    DWC_bool_t        found = FALSE;

    if( !TimerQueue.inited )
    {
        dprintLib( "Timer engine is not initalized.\n" );
        return FALSE;
    }
    if( Timer == INVALID_TIMER )
    {
        dprintLib( "Invalid timer requested.\n" );
        return FALSE;
    }
    else if( Timer == INFINITE_TIMER )
    {
        dprintLib( "Infinite timeout can not be alarmed.\n" );
        return FALSE;
    }
    /* It's guard *ONLY* against one thread loop:
     *  alarm -> lock(); call -> user_proc() -> and user call us recursively lock()
     */
    if( SelfLocking() )
    {
        return FALSE;
    }
    if( !LockQueue( &TimerQueue ) )
    {
        dprintLib( "Can not lock timer engine.\n" );
        return FALSE;
    }
    TAILQ_FOREACH( tmpTimer, &TimerQueue.active_head, active )
    {
        if( tmpTimer == Timer )
        {
            found = TRUE;
            break;
        }
    }
    if( !found )
    {
        dprintLib( "Invalid timer: not existing or already alarmed.\n" );
        if( !UnLockQueue( &TimerQueue ) )
        {
            dprintLib( "Can not unlock timer engine.\n" );
        }
        return FALSE;
    }
    AlarmTimer( Timer );
    if( !UnLockQueue( &TimerQueue ) )
    {
        dprintLib( "Can not unlock timer engine.\n" );
        return FALSE;
    }
    return TRUE;
}

/* --------------------------------------------------------------------- */
DWC_bool_t ExpiredTimeout( DWC_timer_hnd_t Timer )
{
    DWC_timer_hnd_t   tmpTimer;
    DWC_bool_t        retval = FALSE, found = FALSE;

    if( !TimerQueue.inited )
    {
        dprintLib( "Timer engine is not initalized.\n" );
        return TRUE;
    }
    if( Timer == INVALID_TIMER )
    {
        dprintLib( "Invalid timer requested.\n" );
        return TRUE;
    }
    else if( Timer == INFINITE_TIMER )
    {
        return FALSE;
    }
    /* It's guard *ONLY* against one thread loop:
     *  alarm -> lock(); call -> user_proc() -> and user call us recursively lock()
     */
    if( SelfLocking() )
    {
        return TRUE;
    }
    if( !LockQueue( &TimerQueue ) )
    {
        dprintLib( "Can not lock timer engine.\n" );
        return TRUE;
    }
    TAILQ_FOREACH( tmpTimer, &TimerQueue.all_head, all )
    {
        if( tmpTimer == Timer )
        {
            found = TRUE;
            retval = Timer->alarmed;
            break;
        }
    }
    if( !UnLockQueue( &TimerQueue ) )
    {
        dprintLib( "Can not unlock timer engine.\n" );
        return TRUE;
    }
    if( !found )
    {
        dprintLib( "Invalid timer (%p) for test: not existing or already deleted.\n", Timer );
        return TRUE;
    }
    return retval;
}

/* --------------------------------------------------------------------- */
DWC_bool_t ClearTimeout( DWC_timer_hnd_t Timer )
{
    DWC_timer_hnd_t   tmpTimer;
    DWC_bool_t        found = FALSE;

    if( !TimerQueue.inited )
    {
        dprintLib( "Timer engine is not initalized.\n" );
        return FALSE;
    }
    if( Timer == INVALID_TIMER )
    {
        dprintLib( "Invalid timer requested.\n" );
        return FALSE;
    }
    else if( Timer == INFINITE_TIMER )
    {
        return TRUE;
    }
    /* It's guard *ONLY* against on thread loop:
     *  alarm -> lock(); call -> user_proc() -> and user call us recursively lock()
     */
    if( SelfLocking() )
    {
        return FALSE;
    }
    if( !LockQueue( &TimerQueue ) )
    {
        dprintLib( "Can not lock timer engine.\n" );
        return FALSE;
    }
    TAILQ_FOREACH( tmpTimer, &TimerQueue.all_head, all )
    {
        if( tmpTimer == Timer )
        {
            found = TRUE;
            break;
        }
    }
    if( !found )
    {
        if( !UnLockQueue( &TimerQueue ) )
        {
            dprintLib( "Can not unlock timer engine.\n" );
        }
        dprintLib( "Invalid timer (%p) for clearing: not existing.\n", Timer );
        return FALSE;
    }
    RemoveTimer( Timer, TRUE );
    if( !UnLockQueue( &TimerQueue ) )
    {
        dprintLib( "Can not unlock timer engine.\n" );
        return FALSE;
    }
    return TRUE;
}

#endif /* __DWC_TIMO_C */
