#ifndef __DWC_UNIX_C
#define __DWC_UNIX_C

#if defined(__UNIX_DW) && !defined(EMX_DW)

#ifndef lint
static const char rcsid[] = "$Id: dwc_unix.c,v 1.20 2001/10/08 15:48:52 cvser Exp $";
#endif

#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#include "dwc_lib.h"
#include "dwc_unix.h"
#include "dwc_err.h"
#include "dwc_macro.h"

/* -------------------------------------------------------------------- */
DWC_error_code_t init_unix_sockets ( void )
{
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t dinit_unix_sockets ( void )
{
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t addr_unix_op ( DWC_unix_socket_addr_t* addr, DWC_cmd_t cmd, DWC_string_t name )
{
    DWC_bool_t           used = FALSE;
    DWC_error_code_t     error;

    PARAM_ASSERT( addr );
    PARAM_ASSERT( name );

    if( cmd == DWC_IPC_CMD_SETADDR )
    {
        used = TRUE;
        if( name[0] == '\0' )
        {
            error = DWC_BADPARAMS;
            goto error_exit;
        }
        strncpy( addr->sun_path, name, sizeof(addr->sun_path) );
        addr->sun_path[sizeof(addr->sun_path)-1] = NULLCHAR;
    }
    if( !used )
    {
        error = DWC_BADPARAMS;
        goto error_exit;
    }
    error = DWC_NOERROR;
  error_exit:
    return error;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t create_unix_socket( DWC_unix_socket_t *sock, DWC_bool_t nonblocking )
{
    int   opt;

    PARAM_ASSERT( sock );

    *sock = socket( PF_UNIX, SOCK_STREAM, 0 );
    if( (*sock) == INVALID_UNIX_SOCKET )
    {
        dprintLib( "socket() error on unix socket - %s.\n", stdlib_error(errno) );
        return DWC_IPCSYSTEM;
    }
    if( nonblocking )
    {
        opt = fcntl( (*sock), F_GETFL, 0 );
        if( fcntl( (*sock), F_SETFL, opt | O_NONBLOCK ) == -1 )
        {
            dprintLib("Can not set nonblocking mode for UNIX socket.\n");
            close( (*sock) );
            *sock = INVALID_UNIX_SOCKET;
            return DWC_IPCSYSTEM;
        }
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t bind_unix_socket ( DWC_unix_socket_t sock, DWC_unix_socket_addr_t* addr )
{
    DWC_socklen_t len = sizeof(DWC_unix_socket_addr_t);

    if( bind( sock, (struct sockaddr*) addr, len ) < 0 )
    {
        dprintLib("bind() error on unix socket - %s.\n", stdlib_error(errno) );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t listen_unix_socket ( DWC_unix_socket_t sock, int backlog )
{
    if( listen( sock, backlog ) != 0 )
    {
        dprintLib("listen() error on unix socket - %s.\n", stdlib_error(errno) );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t accept_unix_socket ( DWC_unix_socket_t sock, DWC_unix_socket_addr_t *addr, DWC_unix_socket_t *newsock )
{
    DWC_socklen_t len = sizeof(DWC_unix_socket_addr_t);

    *newsock = accept( sock, (struct sockaddr*)  addr, &len );
    if( (*newsock) == INVALID_UNIX_SOCKET )
    {
        dprintLib("accept() error on unix socket - %s.\n", stdlib_error(errno) );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t connect_unix_socket ( DWC_unix_socket_t sock, DWC_unix_socket_addr_t *addr, DWC_timeout_t timeout )
{
    DWC_error_code_t  err;
    int               how, retval;

    retval = connect( sock, (struct sockaddr*) addr, sizeof(DWC_unix_socket_addr_t) );
    if( retval < 0 )
    {
        if( retval == EINPROGRESS )
        {
            how = DWC_IPC_WAIT_EXCP | DWC_IPC_WAIT_SEND;
            err = wait_unix_socket( sock, &how, timeout );
            if( err == DWC_NOERROR  )
            {
                if( how & DWC_IPC_WAIT_SEND )
                {
                    return DWC_NOERROR;
                }
                dprintLib("Connection failed.\n");
                err = DWC_IPCSYSTEM;
            }
            return err;
        }
        dprintLib("connect() error on unix socket - %s.\n", stdlib_error(errno) );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t send_unix_socket ( DWC_unix_socket_t sock, DWC_pointer_t buf, DWC_length_t *len, DWC_timeout_t timeout )
{
    int sent;

    sent = send( sock, buf, *len, 0 );
    if( sent <= 0 )
    {
        (*len) = 0;
        if( sent == EWOULDBLOCK )
        {
            return DWC_NOERROR;
        }
        dprintLib("send() error on unix socket - %s.\n", stdlib_error(errno) );
        return DWC_IPCSYSTEM;
    }
    (*len) = sent;
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t recv_unix_socket ( DWC_unix_socket_t sock, DWC_pointer_t buf, DWC_length_t *len, DWC_timeout_t timeout )
{
    int readed;

    readed = recv( sock, buf, *len, 0 );
    if( readed < 0 )
    {
        (*len) = 0;
        if( readed == EWOULDBLOCK )
        {
            return DWC_NOERROR;
        }
        dprintLib("recv() error on unix socket - %s.\n", stdlib_error(errno) );
        return DWC_IPCSYSTEM;
    }
    if( readed == 0 )
    {
        dprintLib( "recv() error on unix socket (req=%d) - connection closed by remote side.\n", *len );
        (*len) = 0;
        return DWC_IPCSYSTEM;
    }
    if( readed != (*len) )
    {
        dprintLib( "recv() on unix socket req=%d, readed=%d.\n", *len, readed );
    }
    (*len) = readed;
    return DWC_NOERROR;
}
/* -------------------------------------------------------------------- */
DWC_error_code_t wait_unix_socket( DWC_unix_socket_t sock, int *how, DWC_timeout_t timeout )
{
    fd_set         rd_set, wr_set, ex_set;
    struct timeval timo, *pt;
    int            rd;

    FD_ZERO( &rd_set );
    FD_ZERO( &wr_set );
    FD_ZERO( &ex_set );
    if( (*how & DWC_IPC_WAIT_RECV) == DWC_IPC_WAIT_RECV )
    {
        FD_SET( sock, &rd_set );
    }
    if( (*how & DWC_IPC_WAIT_SEND) == DWC_IPC_WAIT_SEND )
    {
        FD_SET( sock, &wr_set );
    }
    if( (*how & DWC_IPC_WAIT_EXCP) == DWC_IPC_WAIT_EXCP )
    {
        FD_SET( sock, &ex_set );
    }
    *how = 0;
    pt = NULL;
    if( timeout > 0 )
    {
        timo.tv_sec = timeout / 1000000;
        timo.tv_usec = timeout % 1000000;
        pt = &timo;
    }
    rd = select( sock+1, &rd_set, &wr_set, &ex_set, pt );
    if( (rd == 0) && pt )
    {
        return DWC_TIMED;
    }
    if( rd < 0 )
    {
        dprintLib("select() error on TCP socket.\n");
        return DWC_IPCSYSTEM;
    }
    if( FD_ISSET( sock, &rd_set ) )
    {
        *how |= DWC_IPC_WAIT_RECV;
    }
    if( FD_ISSET( sock, &wr_set ) )
    {
        *how |= DWC_IPC_WAIT_SEND;
    }
    if( FD_ISSET( sock, &ex_set ) )
    {
        *how |= DWC_IPC_WAIT_EXCP;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t shutdown_unix_socket ( DWC_unix_socket_t *sock, DWC_unix_socket_addr_t addr, DWC_bool_t linked )
{
    int serror, uerror;

    serror = shutdown( *sock, SHUT_RDWR );
    if( linked )
    {
        uerror = unlink( addr.sun_path );
    }
    else
    {
        uerror = DWC_NOERROR;
    }
    (*sock) = INVALID_UNIX_SOCKET;
    if( serror || uerror )
    {
        dprintLib("shutdown() error on unix socket - %s.\n", stdlib_error(errno) );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t close_unix_socket ( DWC_unix_socket_t *sock, DWC_unix_socket_addr_t addr, DWC_bool_t linked )
{
    int cerror, uerror;

    cerror = close( *sock );
    if( linked )
    {
        uerror = unlink( addr.sun_path );
    }
    else
    {
        uerror = DWC_NOERROR;
    }
    *sock = INVALID_UNIX_SOCKET;
    if( cerror || uerror )
    {
        dprintLib( "close() error on unix socket - %s.\n", stdlib_error(errno) );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------- */
void cancel_unix_call( DWC_pointer_t sock )
{
    dprintLib( "UNIX sockets: syscall cancelled - timeout.\n" );
#if defined(FreeBSD_DW) && defined(__THREAD_BASED_DW)
    if( close( *(DWC_unix_socket_t*)(sock) ) < 0 )
#else /* !(FreeBSD_DW && __THREAD_BASED) */
    if( shutdown( *(DWC_unix_socket_t*)(sock), SHUT_RDWR ) < 0 )
#endif /* FreeBSD_DW */
    {
        dprintLib( "UNIX sockets: cancelling blocking syscall failed - %s.\n", stdlib_error(errno) );
    }
    *(DWC_unix_socket_t*)(sock) = -1;
}

#else /* !UNIX || EMX */

#include "dwc_unix.h"
#include "dwc_err.h"

/* -------------------------------------------------- */
DWC_error_code_t init_unix_sockets ( void )
{
    return DWC_NOTSUPPORTED;  
}

/* -------------------------------------------------- */
DWC_error_code_t dinit_unix_sockets ( void )
{
    return DWC_NOTSUPPORTED;  
}

/* -------------------------------------------------- */
DWC_error_code_t create_unix_socket( DWC_unix_socket_t *sock, DWC_bool_t nonblocking )
{
    return DWC_NOTSUPPORTED;  
}

/* -------------------------------------------------- */
DWC_error_code_t addr_unix_op ( DWC_unix_socket_addr_t* addr, DWC_cmd_t cmd, DWC_string_t name )
{
    return DWC_NOTSUPPORTED;  
}

/* -------------------------------------------------- */
DWC_error_code_t bind_unix_socket ( DWC_unix_socket_t sock, DWC_unix_socket_addr_t* addr )
{
    return DWC_NOTSUPPORTED;  
}

/* -------------------------------------------------- */
DWC_error_code_t listen_unix_socket ( DWC_unix_socket_t sock, int backlog )
{
    return DWC_NOTSUPPORTED;  
}

/* -------------------------------------------------- */
DWC_error_code_t accept_unix_socket ( DWC_unix_socket_t sock, DWC_unix_socket_addr_t *addr, DWC_unix_socket_t *newsock )
{
    return DWC_NOTSUPPORTED;  
}

/* -------------------------------------------------- */
DWC_error_code_t connect_unix_socket ( DWC_unix_socket_t sock, DWC_unix_socket_addr_t *addr, DWC_timeout_t timeout )
{
    return DWC_NOTSUPPORTED;  
}

/* -------------------------------------------------- */
DWC_error_code_t send_unix_socket ( DWC_unix_socket_t sock, DWC_pointer_t buf, DWC_length_t* len_ptr, int options )
{
    return DWC_NOTSUPPORTED;  
}

/* -------------------------------------------------- */
DWC_error_code_t recv_unix_socket ( DWC_unix_socket_t sock, DWC_pointer_t buf, DWC_length_t* len_ptr, int options )
{
    return DWC_NOTSUPPORTED;  
}

/* -------------------------------------------------- */
DWC_error_code_t wait_unix_socket( DWC_unix_socket_t sock, int *how, DWC_timeout_t timeout )
{
    return DWC_NOTSUPPORTED;  
}

/* -------------------------------------------------- */
DWC_error_code_t shutdown_unix_socket ( DWC_unix_socket_t *sock, DWC_unix_socket_addr_t addr, DWC_bool_t linked )
{
    return DWC_NOTSUPPORTED;  
}

/* -------------------------------------------------- */
DWC_error_code_t close_unix_socket ( DWC_unix_socket_t *sock, DWC_unix_socket_addr_t addr, DWC_bool_t linked )
{
    return DWC_NOTSUPPORTED;  
}

/* -------------------------------------------------- */
void cancel_unix_call( DWC_pointer_t sock )
{
}

#endif /* __UNIX_DW && !EMX */

#endif /* __DWC_UNIX_H */
